/**
 * Tomislav Podhraški
 * http://www.justpinegames.com
 */

package  
{
	import com.adobe.utils.AGALMiniAssembler;
	import starling.display.BlendMode;
	import starling.textures.Texture;
    
    import flash.display3D.*;
    import flash.geom.*;
    
    import starling.core.RenderSupport;
    import starling.core.Starling;
    import starling.display.DisplayObject;
    import starling.errors.MissingContextError;
    import starling.events.Event;
    import starling.utils.VertexData;
	
	public class PalettedImage extends DisplayObject
	{
        private static var PROGRAM_NAME:String = "PalettedImage";
		
		private var _texture:Texture;
		private var _palette:Texture;
		
        // vertex data 
        private var _vertexData:VertexData;
        private var _vertexBuffer:VertexBuffer3D;
        
        // index data
        private var _indexData:Vector.<uint>;
        private var _indexBuffer:IndexBuffer3D;
        
        // helper objects (to avoid temporary objects)
        private static var sHelperMatrix:Matrix = new Matrix();
        private static var sRenderAlpha:Vector.<Number> = new <Number>[1.0, 1.0, 1.0, 1.0];
        
        public function PalettedImage(texture:Texture, palette:Texture)
        {
			// we'll discard the alpha pixels, so there is no need for blending
			this.blendMode = BlendMode.NONE;
			
			_texture = texture;
			_palette = palette;
            Starling.current.addEventListener(Event.CONTEXT3D_CREATE, onContextCreated);
			
			updateShape();
        }
		
		private function updateShape():void
		{
            // setup vertex data and prepare shaders
            setupVertices();
            createBuffers();
            registerPrograms();
		}
        
        public override function dispose():void
        {
            Starling.current.removeEventListener(Event.CONTEXT3D_CREATE, onContextCreated);
            
            if (_vertexBuffer) _vertexBuffer.dispose();
            if (_indexBuffer) _indexBuffer.dispose();
            
            super.dispose();
        }
        
        private function onContextCreated(event:Event):void
        {
            // the old context was lost, so we create new buffers and shaders.
            createBuffers();
            registerPrograms();
        }
        
        public override function getBounds(targetSpace:DisplayObject, resultRect:Rectangle=null):Rectangle
        {
			if (!_texture || !_palette) return new Rectangle();

            if (resultRect == null) resultRect = new Rectangle();
            
            var transformationMatrix:Matrix = targetSpace == this ? 
                null : getTransformationMatrix(targetSpace, sHelperMatrix);
            
            return _vertexData.getBounds(transformationMatrix, 0, -1, resultRect);
        }

	
		// Generate the actual geometry
        private function setupVertices():void
        {
			if (!_texture) return;
			
			var frame:Rectangle = _texture.frame;
            var width:Number = frame ? frame.width : _texture.width;
            var height:Number = frame ? frame.height : _texture.height;
			
            _vertexData = new VertexData(4);
			_vertexData.setUniformColor(0xffffff);
			
			_vertexData.setPosition(0, 0, 0);
			_vertexData.setTexCoords(0, 0, 0);
			
			_vertexData.setPosition(1, width, 0);
			_vertexData.setTexCoords(1, 1, 0);
			
			_vertexData.setPosition(2, width, height);
			_vertexData.setTexCoords(2, 1, 1);
			
			_vertexData.setPosition(3, 0, height);
			_vertexData.setTexCoords(3, 0, 1);
			
            _indexData = new Vector.<uint>();
            _indexData.push(0, 1, 2);
            _indexData.push(0, 2, 3);
        }
        
        private function createBuffers():void
        {
			if (!_texture) return;

            var context:Context3D = Starling.context;
            if (context == null) throw new MissingContextError();
            
            if (_vertexBuffer) _vertexBuffer.dispose();
            if (_indexBuffer) _indexBuffer.dispose();
            
            _vertexBuffer = context.createVertexBuffer(_vertexData.numVertices, VertexData.ELEMENTS_PER_VERTEX);
            _vertexBuffer.uploadFromVector(_vertexData.rawData, 0, _vertexData.numVertices);
            
            _indexBuffer = context.createIndexBuffer(_indexData.length);
            _indexBuffer.uploadFromVector(_indexData, 0, _indexData.length);
        }
        
        public override function render(support:RenderSupport, alpha:Number):void
        {
			if (!_texture) return;
			
            // always call this method when you write custom rendering code!
            // it causes all previously batched quads/images to render.
            support.finishQuadBatch();
            
            sRenderAlpha[0] = sRenderAlpha[1] = sRenderAlpha[2] = 1.0;
            sRenderAlpha[3] = alpha * this.alpha;
            
            var context:Context3D = Starling.context;
            if (context == null) throw new MissingContextError();
            
            // apply the current blendmode
            support.applyBlendMode(false);
            
            // activate program (shader) and set the required buffers / constants 
            context.setProgram(Starling.current.getProgram(PROGRAM_NAME));
            context.setVertexBufferAt(0, _vertexBuffer, VertexData.POSITION_OFFSET, Context3DVertexBufferFormat.FLOAT_2); 
            context.setVertexBufferAt(1, _vertexBuffer, VertexData.COLOR_OFFSET,    Context3DVertexBufferFormat.FLOAT_4);
            context.setVertexBufferAt(2, _vertexBuffer, VertexData.TEXCOORD_OFFSET,    Context3DVertexBufferFormat.FLOAT_2);
            context.setProgramConstantsFromMatrix(Context3DProgramType.VERTEX, 0, support.mvpMatrix3D, true);            
            context.setProgramConstantsFromVector(Context3DProgramType.VERTEX, 4, sRenderAlpha, 1);
			
			// threshold value for transparency
			var maximumAlpha:Number = 0.999; 
			context.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 0, new <Number>[maximumAlpha, maximumAlpha, maximumAlpha, maximumAlpha], 1);
			
			// setting multiple textures, context is Stage3D context
			context.setTextureAt(0, _texture.base); // fs0 in shader
 		    context.setTextureAt(1, _palette.base); // fs1 in shader
			
            // draw the image
            context.drawTriangles(_indexBuffer, 0, 2);
            
			// reset textures
			context.setTextureAt(0, null);
			context.setTextureAt(1, null);
			
            // reset buffers
            context.setVertexBufferAt(0, null);
            context.setVertexBufferAt(1, null);
            context.setVertexBufferAt(2, null);
        }
		
        private function registerPrograms():void
        {
            var target:Starling = Starling.current;
            if (target.hasProgram(PROGRAM_NAME)) return; // already registered
            
            // va0 -> position
            // va1 -> color
            // vc0 -> mvpMatrix (4 vectors, vc0 - vc3)
            // vc4 -> alpha
            
			var vertexProgramCode:String =
				"m44 op, va0, vc0 \n" + // 4x4 matrix transform to output space
				"mul v0, va1, vc4 \n" + // multiply color with alpha and pass it to fragment shader
				"mov v1, va2 \n";  // pass texture coordinates to fragment program
				
				
			/////////////////////////////////////////////////////////////////////////////////////
			
			// Fragment program which does most of the work here
				
			var fragmentProgramCode:String =
				"tex ft1, v1, fs0 <2d, linear, mipnone, repeat> \n" + // read value from regular texture (to ft1)
				"sub ft2.x, fc0.x, ft1.x \n" + // subtract the texture value (ft1.x) from alpha threshold (fc0.x defined to be 0.999 in main code)
				"kil ft2.x \n" + // discard fragment if it represents the mask, 'kil' does that if the value is less than 0
				"tex ft2, ft1, fs1 <2d, nearest, mipnone, repeat> \n" + // read color from palette using value from the regular texture (ft1)
				"mul oc, ft2, v0 \n";   // multiply vertex color with palette color and store it to the output
			
				
			/////////////////////////////////////////////////////////////////////////////////////
				
            var vertexProgramAssembler:AGALMiniAssembler = new AGALMiniAssembler();
            vertexProgramAssembler.assemble(Context3DProgramType.VERTEX, vertexProgramCode);
            
            var fragmentProgramAssembler:AGALMiniAssembler = new AGALMiniAssembler();
            fragmentProgramAssembler.assemble(Context3DProgramType.FRAGMENT, fragmentProgramCode);
            
            target.registerProgram(PROGRAM_NAME, vertexProgramAssembler.agalcode, fragmentProgramAssembler.agalcode);
        }
		
		public function get texture():Texture 
		{
			return _texture;
		}
		
		public function set texture(value:Texture):void 
		{
			_texture = value;
			updateShape();
		}
		
		public function get palette():Texture 
		{
			return _palette;
		}
		
		public function set palette(value:Texture):void 
		{
			_palette = value;
			updateShape();
		}

	}

}